﻿using System;
using System.Collections.Generic;
using System.Reflection;
using System.Windows;
using Telerik.Windows.Controls;
using Telerik.Windows.Documents.History;
using Telerik.Windows.Documents.Model;
using Telerik.Windows.Documents.RichTextBoxCommands;
using Telerik.Windows.Documents.RichTextBoxCommands.CommandEventArgs;
using Telerik.Windows.Controls.RichTextBoxUI;
using Telerik.Windows.Documents.Commands;
using System.Diagnostics;
using Telerik.Windows.Documents;
namespace AM.Common.UI
{
	/// <summary>
	/// Handles the attachment of the MultiUndoCommand to a <c>RadRichTextBox</c>.
	/// </summary>
	public class RichTextBoxUndoRibbonUI
	{
		private readonly static DependencyProperty CommandStateChangedHandlerProperty;

		private readonly static DependencyProperty ControlStateChangedHandlerProperty;

		private readonly static DependencyProperty ControlStateChangedInternallyHandlerProperty;

		/// <summary>
		/// The Undo command property
		/// </summary>
		public readonly static DependencyProperty RichTextCommandProperty;

		static RichTextBoxUndoRibbonUI()
		{
			RichTextBoxUndoRibbonUI.CommandStateChangedHandlerProperty = DependencyProperty.RegisterAttached("CommandStateChangedHandler", typeof(object), typeof(RichTextBoxUndoRibbonUI), null);
			RichTextBoxUndoRibbonUI.ControlStateChangedInternallyHandlerProperty = DependencyProperty.RegisterAttached("ControlStateChangedInternallyHandler", typeof(object), typeof(RichTextBoxUndoRibbonUI), null);
			RichTextBoxUndoRibbonUI.ControlStateChangedHandlerProperty = DependencyProperty.RegisterAttached("ControlStateChangedHandler", typeof(object), typeof(RichTextBoxUndoRibbonUI), null);
			RichTextBoxUndoRibbonUI.RichTextCommandProperty = DependencyProperty.RegisterAttached("RichTextCommand", typeof(RichTextBoxCommandBase), typeof(RichTextBoxUndoRibbonUI), new PropertyMetadata(RichTextBoxUndoRibbonUI.OnRichTextBoxCommandChanged));
		}

		private static void ClearAttachedProperties(DependencyObject obj)
		{
			obj.ClearValue(RichTextBoxUndoRibbonUI.ControlStateChangedHandlerProperty);
			obj.ClearValue(RichTextBoxUndoRibbonUI.CommandStateChangedHandlerProperty);
			obj.ClearValue(RichTextBoxUndoRibbonUI.ControlStateChangedInternallyHandlerProperty);
		}



		/// <summary>
		/// Initializes a new instance of the <see cref="RichTextBoxUndoRibbonUI"/> class.
		/// </summary>
		public RichTextBoxUndoRibbonUI()
		{
		}


		private static void SetCommandStateChangedHandler(DependencyObject obj, object value)
		{
			obj.SetValue(RichTextBoxUndoRibbonUI.CommandStateChangedHandlerProperty, value);
		}

		private static object GetCommandStateChangedHandler(DependencyObject obj)
		{
			return obj.GetValue(RichTextBoxUndoRibbonUI.CommandStateChangedHandlerProperty);
		}

		private static void SetControlStateChangedHandler(DependencyObject obj, object value)
		{
			obj.SetValue(RichTextBoxUndoRibbonUI.ControlStateChangedHandlerProperty, value);
		}

		private static object GetControlStateChangedHandler(DependencyObject obj)
		{
			return obj.GetValue(RichTextBoxUndoRibbonUI.ControlStateChangedHandlerProperty);
		}

		/// <summary>
		/// Gets the rich text command.
		/// </summary>
		/// <param name="obj">The DependencyObject</param>
		/// <returns>RichTextBoxCommandBase</returns>
		public static RichTextBoxCommandBase GetRichTextCommand(DependencyObject obj)
		{
			return (RichTextBoxCommandBase)obj.GetValue(RichTextBoxUndoRibbonUI.RichTextCommandProperty);
		}

		/// <summary>
		/// Sets the rich text command.
		/// </summary>
		/// <param name="obj">The DependencyObject</param>
		/// <param name="value">RichTextBoxCommandBase</param>
		public static void SetRichTextCommand(DependencyObject obj, RichTextBoxCommandBase value)
		{
			obj.SetValue(RichTextBoxUndoRibbonUI.RichTextCommandProperty, value);
		}

		private static void OnRichTextBoxCommandChanged(DependencyObject o, DependencyPropertyChangedEventArgs e)
		{
			if (o as MultipleUndoControl != null && (e.NewValue as MultiUndoCommand != null || e.OldValue as MultiUndoCommand != null))
			{
				AttachMultiUndoCommand((MultipleUndoControl)o, e.OldValue as MultiUndoCommand, e.NewValue as MultiUndoCommand);
			}
			if (e.OldValue != null && e.NewValue == null)
			{
				ClearAttachedProperties(o);
			}
		}

		class UndoMergeParams
		{
			/// <summary>
			/// will contain modified list of actions with Typing actions merged
			/// </summary>
			public List<HistoryAction> actionList;

			/// <summary>
			/// will contain the entries for the control's dropdown list 
			/// </summary>
			public List<MultipleUndoPoint> mupList;

			public MultipleUndoPoint mup;

			// actions are lifo, so the last will be the first encountered and the first will be the last
			public HistoryCommandAction firstTypingAction;
			public HistoryCommandAction lastTypingAction;

			public bool inTyping;
			public int inx;

			public UndoMergeParams()
			{
				actionList = new List<HistoryAction>();
				mupList = new List<MultipleUndoPoint>();
				mup = null;
				firstTypingAction = null;
				lastTypingAction = null;
				inTyping = false;
				inx = 0;
			}
		};

		private static void AttachMultiUndoCommand(MultipleUndoControl multipleUndoControl, MultiUndoCommand oldCommand, MultiUndoCommand newCommand)
		{
			
			EventHandler eventHandler = null;
			EventHandler<MultipleUndoExecutedEventArgs> eventHandler1 = null;
			if (oldCommand != null)
			{
				EventHandler oldHandler = GetCommandStateChangedHandler(multipleUndoControl) as EventHandler;
				if (oldHandler != null)
				{
					oldCommand.DocumentHistoryChanged -= oldHandler;
				}
				EventHandler<MultipleUndoExecutedEventArgs> oldmultipleUndoHandler = GetControlStateChangedHandler(multipleUndoControl) as EventHandler<MultipleUndoExecutedEventArgs>;
				if (oldmultipleUndoHandler != null)
				{
					multipleUndoControl.MultipleUndoTriggered -= oldmultipleUndoHandler;
				}
			}
			if (newCommand != null)
			{
				eventHandler = (object s, EventArgs e) =>
				{
					var undoStack = GetUndoStack(newCommand);

					var undoMergeParams = new UndoMergeParams();
						
					foreach (HistoryCommandAction a in undoStack)
					{
						if (((undoMergeParams.inTyping) && (a.Command.Name.CompareTo("Typing") != 0)))
						{
							FinishTypingMerge(undoMergeParams);
						}

						if ((!undoMergeParams.inTyping) && (a.Command.Name.CompareTo("Typing") == 0))
						{
							StartTypingMerge(undoMergeParams, a);
						}

						if (undoMergeParams.inTyping)
						{
							UpdateMupDisplayText(undoMergeParams, a);
						}
						else
						{
							AddNonTypingAction(undoMergeParams, a);
						}

						if (undoMergeParams.mupList.Count >= 20)
						{
							break;
						}
					}

					if (undoMergeParams.inTyping)
					{
						FinishTypingMerge(undoMergeParams);
					}
						
					// rebuild the undo stack from list of merged actions
					undoStack.Clear();
					foreach (HistoryAction a in undoMergeParams.actionList)
					{
						undoStack.Push(a);
					}

					if (!multipleUndoControl.IsTemplateApplied)
					{
						multipleUndoControl.ItemsCollection = undoMergeParams.mupList;
					}
					else
					{
						multipleUndoControl.SetItemsControlSource(undoMergeParams.mupList);
					}
				};
				
				EventHandler newHandler = eventHandler;
				SetCommandStateChangedHandler(multipleUndoControl, newHandler);
				newCommand.DocumentHistoryChanged += newHandler;
				newHandler(newCommand, EventArgs.Empty);
				if (eventHandler1 == null)
				{
					eventHandler1 = (object s, MultipleUndoExecutedEventArgs e) => 
					{
						newCommand.Execute(e.UndoStackIndex);
					};
				}
				EventHandler<MultipleUndoExecutedEventArgs> multipleUndoHandler = eventHandler1;
				multipleUndoControl.MultipleUndoTriggered += multipleUndoHandler;
				SetControlStateChangedHandler(multipleUndoControl, multipleUndoHandler);
			}
		}

		private static Stack<HistoryAction> GetUndoStack(MultiUndoCommand newCommand)
		{
			Type historyType = newCommand.AssociatedRichTextBox.Document.History.GetType();
			return historyType.InvokeMember("undoStack", BindingFlags.NonPublic | BindingFlags.Instance | BindingFlags.GetField, null,
				newCommand.AssociatedRichTextBox.Document.History, null) as Stack<HistoryAction>;
		}

		private static void StartTypingMerge(UndoMergeParams undoMergeParams, HistoryCommandAction a)
		{
			undoMergeParams.lastTypingAction = a;
			undoMergeParams.mup = new MultipleUndoPoint();
			undoMergeParams.inTyping = true;
		}

		private static void UpdateMupDisplayText(UndoMergeParams undoMergeParams, HistoryCommandAction a)
		{
			undoMergeParams.firstTypingAction = a;
			
			var ctx = undoMergeParams.firstTypingAction.CommandContext;
			if (ctx != null)
			{
				try
				{
					Type commandContextType = ctx.GetType();
					String text = commandContextType.InvokeMember("text", BindingFlags.NonPublic | BindingFlags.Instance | BindingFlags.GetField, null,
						ctx, null) as String;

					if (!String.IsNullOrEmpty(text))
					{
						undoMergeParams.mup.DisplayText = text + undoMergeParams.mup.DisplayText;
					}

				}
				catch
				{
					// will throw exception if the context is not an InsertTextCommandContext,
					// so just eat it.
				}
			}
		}

		private static void AddNonTypingAction(UndoMergeParams undoMergeParams, HistoryCommandAction a)
		{
			undoMergeParams.mup = new MultipleUndoPoint();
			undoMergeParams.mup.DisplayText = a.Command.Name;
			undoMergeParams.firstTypingAction = a;
			UpdateLists(undoMergeParams);
		}

		private static void FinishTypingMerge(UndoMergeParams undoMergeParams)
		{
			Type lastActionType = undoMergeParams.lastTypingAction.CommandContext.GetType();
			Type firstActionType = undoMergeParams.firstTypingAction.CommandContext.GetType();

			try
			{
				// get the end position of the last action and assign it to the end position of the first action
				DocumentPosition endPos = lastActionType.BaseType.InvokeMember("EndPosition", BindingFlags.NonPublic | BindingFlags.Public | BindingFlags.Instance | BindingFlags.GetField | BindingFlags.GetProperty, null,
					undoMergeParams.lastTypingAction.CommandContext, null) as DocumentPosition;

				firstActionType.BaseType.InvokeMember("EndPosition", BindingFlags.NonPublic | BindingFlags.Public | BindingFlags.Instance | BindingFlags.SetField | BindingFlags.SetProperty, null,
					undoMergeParams.firstTypingAction.CommandContext, new object[] { endPos });

				// initialize the text and displayText of the
				firstActionType.InvokeMember("text", BindingFlags.NonPublic | BindingFlags.Instance | BindingFlags.SetField, null,
					undoMergeParams.firstTypingAction.CommandContext, new object[] { undoMergeParams.mup.DisplayText });

				undoMergeParams.mup.DisplayText = "Typing \"" + undoMergeParams.mup.DisplayText;
				if (undoMergeParams.mup.DisplayText.Length > 49)
				{
					undoMergeParams.mup.DisplayText = undoMergeParams.mup.DisplayText.Substring(0, 47) + "...";
				}
				else
				{
					undoMergeParams.mup.DisplayText += "\"";
				}

				firstActionType.InvokeMember("displayText", BindingFlags.NonPublic | BindingFlags.Instance | BindingFlags.SetField, null,
					undoMergeParams.firstTypingAction.CommandContext, new object[] { undoMergeParams.mup.DisplayText });

				UpdateLists(undoMergeParams);
			}
			catch (System.Exception ex)
			{
				Debug.WriteLine(ex.Message);
			}

		}

		private static void UpdateLists(UndoMergeParams undoMergeParams)
		{
			undoMergeParams.mup.UndoStackIndex = undoMergeParams.inx++;
			undoMergeParams.mupList.Add(undoMergeParams.mup);
			undoMergeParams.actionList.Insert(0, undoMergeParams.firstTypingAction);
			undoMergeParams.inTyping = false;
		}

	}

}
